#!/usr/bin/env python
# -*- coding: utf-8 -*-
# make score-thai html from yaml. by muri 2006/10/17

import sys, os, getopt, codecs, re, types, yaml
from xml.sax.saxutils import escape
from scorethai_yaml import Song
from tagdata import TagData, CellBuf

# options
opt_verbose = opt_check_format = opt_no_output = opt_dump = False

def msgf(s, *args):
    print >>sys.stderr, str(s) % args
    
class Format(TagData):
    reFlagOn = re.compile(r'^(\S)<')
    reFlagOff = re.compile(r'^>(\S)')
    reMark = re.compile(r'^\*(\S*)')
    cur_flags = [] # current flags
    
    def __init__(self):
        self.clear_flags()
        
    def clear_flags(self):
        self.cur_flags = []
        
    def check_flags(self):
        if self.cur_flags:
            msgf("* Flags not closed: %s", ','.join(self.cur_flags))
    
    def check_tags(self):
        format_keys = ('page', 'table', 'line', 'cell', 'text', 'mark')
        r = True
        for k in format_keys:
            if k not in self:
                msgf('* format[%r] must be defined', k)
                r = False
            lst = self[k]
            if type(lst) != list or len(lst) != 2:
                msgf('* format[%r] must has [open-tag, close-tag]', k)
                r = False
            for i in lst:
                if type(i) not in [str, unicode]:
                    msgf('* format[%r] must be list of string: %r', k, i)
        return r
        
    def tag_flags(self, flags, isopen):
        """Produce open or close tag for flags.
        flags: flag chars.
        isopen: True for open tag, False for close tag.
        """
        r = u''
        for c in flags:
            if c in self:              # defined flag
                i = int(not isopen)
                r += self[c][i]
            else:                       # undefined flag
                msgf('* Undefined flag: %r', c)
        return r
        
    def tag_chord(self, src):
        """Process src string as one chord, return tagged string."""
        r = u''
        while src:
            m = self.reMark.match(src)
            if m:
                mark = m.group(1)
                src = src[m.end():]
                r += self.tag_data('mark', mark)
                continue
            m = self.reFlagOn.match(src)
            if m:
                c = m.group(1)
                src = src[m.end():]
                if c not in self.cur_flags:
                    r += self.tag_flags(c, True)
                    if c in self:
                        self.cur_flags.append(c)
                else:
                    msgf('* Duplicated flag: %r', c)
                continue
            m = self.reFlagOff.match(src)
            if m:
                c = m.group(1)
                src = src[m.end():]
                if c in self.cur_flags:
                    r += self.tag_flags(c, False)
                    if c in self:
                        self.cur_flags.remove(c)
                else:
                    msgf('* Unmatched flag: %r', c)
                continue
            r += src[0] # add one note
            src = src[1:] # next notes
        return r
        
    def tag_cell(self, s):
        """Make tagged cell.
        s: String to put in cell.
        """
        # begin with open tag of current flags, make tagged chords, close with current flags
        s1 = self.tag_flags(self.cur_flags, True)
        chords = [self.tag_chord(sc) for sc in s.split()]
        s2 = self.tag_flags(self.cur_flags, False)
        return self.tag_data('cell', s1 + ' '.join(chords) + s2)
        
    def tag_song(self, song):
        """Make Song to formatted table, return tagged string.
        song: Song
        """
        self.clear_flags()
        self.datadict['song'] = song.title
        columns = song.columns or 8
        self.datadict['columns'] = str(columns)
        lines = CellBuf(columns, self.tag_data('cell'))
        bodies = song.parse()
        for body in bodies:
            if body.typ == body.COLUMNS:
                lines.flush()
                columns = body.val
                self.datadict['columns'] = str(columns)
                lines.setcol(columns)
            elif body.typ == body.TEXT:
                lines.add_line(self.tag_data('text', body.val))
            elif body.typ == body.NEWLINE:
                lines.flush()
            else: # cell, make it tagged.
                lines.add_cell(self.tag_cell(body.val))
        self.check_flags()
        lines.flush()
        s = ''.join([self.tag_data('line', ''.join(sl)) for sl in lines])
        return self.tag_data('table', s)
        
class Data(dict):
    """Data dict for page formatting.
    """
    def update(self, dic):
        """Update data by dict dic.
        """
        msgf(' data keys: %s', ', '.join([str(k) for k in dic]))
        if type(dic) != dict:
            msgf('* data value must be mapping: %r', dic)
            sys.exit(1)
        for k, v in dic.iteritems():
            if type(v) not in (str, unicode):
                msgf('* data value must be string: %r: %r', k, v)
                sys.exit(1)
        dict.update(self, dic)
        
encoding = 'utf-8'

data_default = { 'title': 'score', 'css': 'scorethai.css', 'output': None }
data = Data(data_default)
format = Format()
format.setdata(data)

def output(src):
    """Output tag formatted string to file.
    src: string to write.
    """
    name = data['output']
    if opt_no_output:
        return
    if name:
        msgf('  output to: %r', name)
        f = codecs.open(name, encoding=encoding, mode='w')
        f.write(src)
        f.close()
        data['output'] = None
    else:
        sys.stdout.write(src)

def page(parm, songs):
    """Command 'page'. make score tables for each songs. then output as a page.
    parm: song_id or [ song_id,... ].
    songs: dict { title: Song }.
    """
    msgf(' page: %r', parm)
    if not parm:
        return
    if type(parm) != list:
        parm = [parm]
    rs = ''
    for key in parm:
        msgf('  song: %r', key)
        if key not in songs:
            msgf('* song not found: %r', key)
            sys.exit(1)
        rs += format.tag_song(songs[key])
    output(format.tag_data('page', rs))

def load(f, songs=None):
    """Load and parse yaml source.
    f: input string or file object.
    """
    global opt_verbose
    
    if songs == None:
        songs = {}  # dict { title: Song }
    
    try:
        lst = [i for i in yaml.load_all(f)]
    except yaml.YAMLError, exc:
        msgf("%s\n in object: %r", exc, f)
        sys.exit(1)
    
    if opt_verbose:
        msgf('yaml: %r', lst)
    if opt_dump:
        print 'dump:', f
        for i in lst:
            print yaml.dump(i, default_flow_style=False, width=60)
        
    for y in lst:
        if type(y) == Song:
            songs[y.id] = y
        elif type(y) == dict:
            for (k, v) in y.iteritems():
                if k == 'format':
                    format.update(v)
                elif k == 'data':
                    data.update(v)
                else:
                    msgf('* Unknown key in block: %r : %s', k, v)
        elif type(y) == list:
            for v in y:
                if type(v) == dict:
                    for (cmd, parm) in v.iteritems():
                        if cmd == 'load':
                            if type(parm) != list:
                                parm = [parm]
                            for s in parm:
                                msgf('load: %s', s)
                                f = open(s)
                                load(f, songs)
                                f.close()
                        elif cmd == 'format':
                            format.update(parm)
                        elif cmd == 'data':
                            data.update(parm)
                        elif cmd == 'page':
                            page(parm, songs)
                        else:
                            msgf('* Unknown command in command-list:\n  %r: %r', cmd, parm)
                            sys.exit(1)
                else:
                    msgf('* Unknown block: %r', v)
        else:
            msgf('* Unknown block: %r', y)
        
def usage():
    print >>sys.stderr, '''usage:
prog [option] input-file..
 -h    show  help.
 -d    dump yaml after loads.
 -n    no output.
 -c    check format (also no output).
 -v    verbose.
'''
    sys.exit(2)
    
def main():
    global opt_verbose, opt_check_format, opt_no_output, opt_dump
    
    sys.stdin = codecs.getreader(encoding)(sys.stdin)
    sys.stdout = codecs.getwriter(encoding)(sys.stdout)
    sys.stderr = codecs.getwriter(encoding)(sys.stderr)
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvcdn")
    except getopt.GetoptError:
        usage()
        
    for o, a in opts:
        if o == '-v':
            opt_verbose = True
        elif o == '-c':
            opt_check_format = opt_no_output = True
        elif o == '-n':
            opt_no_output = True
        elif o == '-d':
            opt_dump = True
        else:
            usage()
    
    if args:
        for arg in args:
            msgf("Reading file: %s", arg)
            f = codecs.open(arg, encoding=encoding, mode='r')
            load(f)
            f.close()
    else:
        msgf("Reading stdin:")
        load(sys.stdin)

    if opt_check_format:
        format.check_tags()
        print 'format = {\n' + ''.join(['  %s: %s\n' % (k, repr(v)) for k, v in format.iteritems()]) + '}\n'
    
if __name__=="__main__":
    main()
